disp('compiling ...');
mex -I./../eigen                           ...
    -I./../include                              ...
    cppoptlib.cxx  CXXFLAGS="\$CXXFLAGS          ...
    -std=c++11" COPTIMFLAGS="\$COPTIMFLAGS ...
    -O2"                                   ...
    -output cppoptlib;
